@@ -15,15 +15,13 @@
  11/2004:
  add probing for some features (priority)
 
+ 09/2008:
+ change affinity calls to new cpu_set_t API
+
+
  Born in the need of querying and setting SCHED_* policies.
  All output, even errors, go to STDOUT to ease piping.
 
- - Setting CPU-affinity is now supported, too.
-
- - schedtool can now be used to exec a new process with specific parameters.
-
- - nice functionality is incorporated.
-
 
  Content:
 
@@ -35,6 +33,8 @@
 
  */
 
+#define _GNU_SOURCE
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -46,33 +46,9 @@
 #include <unistd.h>
 #include <stdint.h>
 
-/* this gets us the list of syscalls */
-#include <linux/unistd.h>
-
 #include "error.h"
 #include "util.h"
 
-/*
- CPU-affinity stuff
- have an hack to include support for a newer kernel with older headers
- check if support is really there
-
- It's time for autoconf, baby.
- */
-
-/* provide support for the syscalls even if the libc doesn't know about it */
-#ifdef HAVE_AFFINITY
-
-#ifdef AFFINITY_HACK
-#include "syscall_magic.h"
-#endif
-
-#ifndef __NR_sched_getaffinity
-#error You tried to build with support for affinity, but your system/headers are not ready.
-#error Please see the file INSTALL for more information.
-#endif
-
-#endif
 
 /* various operation modes: print/set/affinity/fork */
 #define MODE_NOTHING	0x0
@@ -81,7 +57,7 @@
 #define MODE_AFFINITY	0x4
 #define MODE_EXEC	0x8
 #define MODE_NICE       0x10
-#define VERSION "1.2.10"
+#define VERSION "1.3.0"
 
 /*
  constants are from the O(1)-sched kernel's include/sched.h
@@ -106,6 +82,13 @@
 #define CHECK_RANGE_POLICY(p) (p <= SCHED_MAX && p >= SCHED_MIN)
 #define CHECK_RANGE_NICE(n) (n <= 20 && n >= -20)
 
+/*
+ 4 bits of CPU_SETSIZE will naturally reduce to 1 char, so
+ we'd only need [CPU_SETSIZE / 4 + 1]
+ to be sure leave a bit of room and only do CPU_SETSIZE / 2
+ */
+#define CPUSET_HEXSTRING(name) char name[CPU_SETSIZE / 2]
+
 char *TAB[] = {
 	"N: SCHED_NORMAL",
 	"F: SCHED_FIFO",
@@ -123,7 +106,7 @@ struct engine_s {
 	int policy;
 	int prio;
         int nice;
-	unsigned long aff_mask;
+	cpu_set_t aff_mask;
 
 	/* # of args when going in PID-mode */
 	int n;
@@ -133,8 +116,12 @@ struct engine_s {
 
 int engine(struct engine_s *e);
 int set_process(pid_t pid, int policy, int prio);
-unsigned long parse_affinity(char *arg);
-int set_affinity(pid_t pid, unsigned long mask);
+static inline int val_to_char(int v);
+static char * cpuset_to_str(cpu_set_t *mask, char *str);
+static inline int char_to_val(int c);
+static int str_to_cpuset(cpu_set_t *mask, const char* str);
+int parse_affinity(cpu_set_t *, char *arg);
+int set_affinity(pid_t pid, cpu_set_t *mask);
 int set_niceness(pid_t pid, int nice);
 void probe_sched_features();
 void get_prio_min_max(int policy, int *min, int *max);
@@ -158,11 +145,10 @@ int main(int ac, char **dc)
 	int policy=-1, nice=10, prio=0, mode=MODE_NOTHING;
 
 	/*
-	 default aff_mask to 0xFF..FF== all CPUs
-         unsigned long has not a defined number of bits on all arches
-	 so: use the biggest type and cast it down so we should stay portable
+	 aff_mask: zero it out
 	 */
-	unsigned long aff_mask=(unsigned long)UINT64_MAX;
+	cpu_set_t aff_mask;
+        CPU_ZERO(&aff_mask);
 
         /* for getopt() */
 	int c;
@@ -211,14 +197,9 @@ int main(int ac, char **dc)
 			mode |= MODE_SETPOLICY;
 			break;
 		case 'a':
-#ifdef HAVE_AFFINITY
 			mode |= MODE_AFFINITY;
-			aff_mask=parse_affinity(optarg);
+			parse_affinity(&aff_mask, optarg);
                         break;
-#else
-			printf("ERROR: compile-time option CPU-affinity is not supported\n");
-                        return(-1);
-#endif
 		case 'n':
                         mode |= MODE_NICE;
 			nice=atoi(optarg);
@@ -336,12 +317,15 @@ int engine(struct engine_s *e)
 	int i;
 
 #ifdef DEBUG
-	printf("Dumping mode: 0x%x\n", e->mode);
-	printf("Dumping affinity: 0x%x\n", e->aff_mask);
-	printf("We have %d args to do\n", e->n);
-	for(i=0;i < e->n; i++) {
-		printf("Dump arg %d: %s\n", i, e->args[i]);
-	}
+	do {
+		CPUSET_HEXSTRING(tmpaff);
+		printf("Dumping mode: 0x%x\n", e->mode);
+		printf("Dumping affinity: 0x%s\n", cpuset_to_str(&(e->aff_mask), tmpaff));
+		printf("We have %d args to do\n", e->n);
+		for(i=0;i < e->n; i++) {
+			printf("Dump arg %d: %s\n", i, e->args[i]);
+		}
+	} while(0);
 #endif
 
 	/*
@@ -351,6 +335,10 @@ int engine(struct engine_s *e)
 	for(i=0; i < e->n; i++) {
 
 		int pid, tmpret=0;
+		cpu_set_t affi;
+
+		CPU_ZERO(&affi);
+                CPU_SET(0, &affi);
 
                 /* if in MODE_EXEC skip check for PIDs */
 		if(mode_set(e->mode, MODE_EXEC)) {
@@ -393,9 +381,8 @@ int engine(struct engine_s *e)
 
 		}
 
-#ifdef HAVE_AFFINITY
 		if(mode_set(e->mode, MODE_AFFINITY)) {
-			tmpret=set_affinity(pid, e->aff_mask);
+			tmpret=set_affinity(pid, &(e->aff_mask));
 			ret += tmpret;
 
 			if(tmpret) {
@@ -403,7 +390,6 @@ int engine(struct engine_s *e)
 			}
 
 		}
-#endif
 
 		/* and print process info when set, too */
 		if(mode_set(e->mode, MODE_PRINT)) {
@@ -455,29 +441,127 @@ int set_process(pid_t pid, int policy, int prio)
 }
 
 
-#ifdef HAVE_AFFINITY
-unsigned long parse_affinity(char *arg)
+/*
+ the following functions have been taken from taskset of util-linux
+ (C) 2004 by Robert Love
+ */
+static inline int val_to_char(int v)
 {
-	unsigned long tmp_aff=0;
+	if (v >= 0 && v < 10)
+		return '0' + v;
+	else if (v >= 10 && v < 16)
+		return ('a' - 10) + v;
+	else 
+		return -1;
+}
+
+
+/*
+ str seems to need to hold [CPU_SETSIZE * 7] chars, as in used in taskset
+ however, 4 bits of CPU_SETSIZE will naturally reduce to 1 char
+ (bits "1111" -> char "f") so CPU_SETSIZE / 4 + 1 should be sufficient
+
+ will pad with zeroes to the start, so print the string starting at the
+ returned char *
+ */
+static char * cpuset_to_str(cpu_set_t *mask, char *str)
+{
+	int base;
+	char *ptr = str;
+	char *ret = 0;
+
+	for (base = CPU_SETSIZE - 4; base >= 0; base -= 4) {
+		char val = 0;
+		if (CPU_ISSET(base, mask))
+			val |= 1;
+		if (CPU_ISSET(base + 1, mask))
+			val |= 2;
+		if (CPU_ISSET(base + 2, mask))
+			val |= 4;
+		if (CPU_ISSET(base + 3, mask))
+			val |= 8;
+		if (!ret && val)
+			ret = ptr;
+		*ptr++ = val_to_char(val);
+	}
+	*ptr = 0;
+	return ret ? ret : ptr - 1;
+}
+
+
+static inline int char_to_val(int c)
+{
+	int cl;
+
+	cl = tolower(c);
+	if (c >= '0' && c <= '9')
+		return c - '0';
+	else if (cl >= 'a' && cl <= 'f')
+		return cl + (10 - 'a');
+	else
+		return -1;
+}
+
+
+static int str_to_cpuset(cpu_set_t *mask, const char* str)
+{
+	int len = strlen(str);
+	const char *ptr = str + len - 1;
+	int base = 0;
+
+	/* skip 0x, it's all hex anyway */
+	if(len > 1 && str[0] == '0' && str[1] == 'x') {
+		str += 2;
+	}
+
+	CPU_ZERO(mask);
+	while (ptr >= str) {
+		char val = char_to_val(*ptr);
+		if (val == (char) -1)
+			return -1;
+		if (val & 1)
+			CPU_SET(base, mask);
+		if (val & 2)
+			CPU_SET(base + 1, mask);
+		if (val & 4)
+			CPU_SET(base + 2, mask);
+		if (val & 8)
+			CPU_SET(base + 3, mask);
+		len--;
+		ptr--;
+		base += 4;
+	}
+
+	return 0;
+}
+/* end of functions taken */
+
+
+/* mhm - we need something clever for all that CPU_SET() and CPU_ISSET() stuff */
+int parse_affinity(cpu_set_t *mask, char *arg)
+{
+	cpu_set_t tmp_aff;
 	char *tmp_arg;
-        size_t valid_len;
+	size_t valid_len;
+
+        CPU_ZERO(&tmp_aff);
 
 	if(*arg == '0' && *(arg+1) == 'x') {
-                /* we're in standard hex mode */
-		tmp_aff=strtol(optarg, NULL, 16);
+		/* we're in standard hex mode */
+		str_to_cpuset(&tmp_aff, arg);
 
 	} else if( (valid_len=strspn(arg, "0123456789,.")) ) {
 		/* new list mode: schedtool -a 0,2 -> run on CPU0 and CPU2 */
 
 		/* split on ',' and '.', because '.' is near ',' :) */
 		while((tmp_arg=strsep(&arg, ",."))) {
-                        int tmp_shift;
+                        int tmp_cpu;
 
 			if(isdigit((int)*tmp_arg)) {
-				tmp_shift=atoi(tmp_arg);
-				tmp_aff |= (0x1 << tmp_shift);
+				tmp_cpu=atoi(tmp_arg);
+                                CPU_SET(tmp_cpu, &tmp_aff);
 #ifdef DEBUG
-				printf("tmp_arg: %s -> tmp_shift: 1 << %d, tmp_aff: 0x%x\n", tmp_arg, tmp_shift, tmp_aff);
+				printf("tmp_arg: %s -> tmp_cpu: %d\n", tmp_arg, tmp_cpu);
 #endif
 			}
 		}
@@ -487,27 +571,25 @@ unsigned long parse_affinity(char *arg)
 		exit(1);
 	}
 
-#ifdef DEBUG
-	printf("Affinity result: 0x%x\n", tmp_aff);
-#endif
-	return tmp_aff;
+        *mask=tmp_aff;
+	return 0;
 }
 
 
-int set_affinity(pid_t pid, unsigned long mask)
+int set_affinity(pid_t pid, cpu_set_t *mask)
 {
 	int ret;
+	CPUSET_HEXSTRING(aff_hex);
 
-	if((ret=sched_setaffinity(pid, sizeof(mask), &mask))) {
-		decode_error("could not set PID %d to affinity 0x%x",
+	if((ret=sched_setaffinity(pid, sizeof(cpu_set_t), mask)) == -1) {
+		decode_error("could not set PID %d to affinity 0x%s",
 			     pid,
-                             mask
+			     cpuset_to_str(mask, aff_hex)
 			    );
 		return(ret);
 	}
         return(0);
 }
-#endif
 
 
 int set_niceness(pid_t pid, int nice)
@@ -577,9 +659,14 @@ void print_process(pid_t pid)
 {
 	int policy, nice;
 	struct sched_param p;
-        unsigned long aff_mask=(0-1);
+	cpu_set_t aff_mask;
+	CPUSET_HEXSTRING(aff_mask_hex);
+
+
+	CPU_ZERO(&aff_mask);
 
 	/* strict error checking not needed - it works or not. */
+        errno=0;
 	if( ((policy=sched_getscheduler(pid)) < 0)
 	    || (sched_getparam(pid, &p) < 0)
 	    /* getpriority may successfully return negative values, so errno needs to be checked */
@@ -607,7 +694,6 @@ void print_process(pid_t pid)
 			      );
 		}
 
-#ifdef HAVE_AFFINITY
 		/*
 		 sched_getaffinity() seems to also return (int)4 on 2.6.8+ on x86 when successful.
 		 this goes against the documentation
@@ -619,9 +705,8 @@ void print_process(pid_t pid)
 			 */
                         errno=0;
 		} else {
-			printf(", AFFINITY 0x%lx", aff_mask);
+			printf(", AFFINITY 0x%s", cpuset_to_str(&aff_mask, aff_mask_hex));
 		}
-#endif
 	printf("\n");
 	}
 }
@@ -630,31 +715,27 @@ void print_process(pid_t pid)
 void usage(void)
 {
 	printf(
-	       "get/set scheduling policies - v" VERSION ", GPL'd, NO WARRANTY\n" \
-	       "USAGE: schedtool PIDS                    - query PIDS\n" \
-	       "       schedtool [OPTIONS] PIDS          - set PIDS\n" \
-	       "       schedtool [OPTIONS] -e COMMAND    - exec COMMAND\n" \
+               "get/set scheduling policies - v" VERSION ", GPL'd, NO WARRANTY\n" \
+               "USAGE: schedtool PIDS                    - query PIDS\n" \
+               "       schedtool [OPTIONS] PIDS          - set PIDS\n" \
+               "       schedtool [OPTIONS] -e COMMAND    - exec COMMAND\n" \
                "\n" \
-	       "set scheduling policies:\n" \
-	       "    -N                    for SCHED_NORMAL\n" \
-	       "    -F -p PRIO            for SCHED_FIFO       only as root\n" \
-	       "    -R -p PRIO            for SCHED_RR         only as root\n" \
-	       "    -B                    for SCHED_BATCH\n" \
-	       "    -I -p PRIO            for SCHED_ISO\n" \
-	       "    -D                    for SCHED_IDLEPRIO\n" \
+               "set scheduling policies:\n" \
+               "    -N                    for SCHED_NORMAL\n" \
+               "    -F -p PRIO            for SCHED_FIFO       only as root\n" \
+               "    -R -p PRIO            for SCHED_RR         only as root\n" \
+               "    -B                    for SCHED_BATCH\n" \
+               "    -I -p PRIO            for SCHED_ISO\n" \
+               "    -D                    for SCHED_IDLEPRIO\n" \
                "\n" \
                "    -M POLICY             for manual mode; raw number for POLICY\n" \
-	       "    -p STATIC_PRIORITY    usually 1-99; only for FIFO or RR\n" \
-	       "                          higher numbers means higher priority\n" \
-	       "    -n NICE_LEVEL         set niceness to NICE_LEVEL\n" \
-	      );
-#ifdef HAVE_AFFINITY
-	printf("    -a AFFINITY_MASK      set CPU-affinity to bitmask or list\n\n");
-#endif
-	printf(
-	       "    -e COMMAND [ARGS]     start COMMAND with specified policy/priority\n" \
-	       "    -r                    display priority min/max for each policy\n" \
-	       "    -v                    be verbose\n" \
+               "    -p STATIC_PRIORITY    usually 1-99; only for FIFO or RR\n" \
+               "                          higher numbers means higher priority\n" \
+               "    -n NICE_LEVEL         set niceness to NICE_LEVEL\n" \
+               "    -a AFFINITY_MASK      set CPU-affinity to bitmask or list\n\n" \
+               "    -e COMMAND [ARGS]     start COMMAND with specified policy/priority\n" \
+               "    -r                    display priority min/max for each policy\n" \
+               "    -v                    be verbose\n" \
 	       "\n" \
 	      );
 /*	printf("Parent ");
