@@ -1,3 +1,10 @@
+/*
+ this sticks around for documentation issues only and will be removed in the next
+ release - it documents the direct syscalls for affinity, without going thru glibc
+ */
+
+#if 0
+
 /* 
  I don't know where this was exactly taken from, but I think I found it
  in glibc.
@@ -20,3 +27,5 @@
  */
 _syscall3 (int, sched_setaffinity, pid_t, pid, unsigned int, len, unsigned long *, user_mask_ptr)
 _syscall3 (int, sched_getaffinity, pid_t, pid, unsigned int, len, unsigned long *, user_mask_ptr)
+
+#endif
